/**
 * Local (StringConnection) network system.  Version 1.2.0.
 * This file Copyright (C) 2007-2009,2017 Jeremy D Monin <jeremy@nand.net>.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * The maintainer of this program can be reached at jsettlers@nand.net
 **/
package soc.server.genericServer;

import java.util.Date;

/**
 * StringConnection allows clients and servers to communicate,
 * with no difference between local and actual networked traffic.
 * 
 *<PRE>
 *  1.0.0 - 2007-11-18 - initial release
 *  1.0.1 - 2008-06-28 - add getConnectTime
 *  1.0.2 - 2008-07-30 - no change in this file
 *  1.0.3 - 2008-08-08 - add disconnectSoft, getVersion, setVersion
 *  1.0.4 - 2008-09-04 - add appData
 *  1.0.5 - 2009-05-31 - add isVersionKnown, setVersion(int,bool),
 *                       setVersionTracking, isInputAvailable,
 *                       wantsHideTimeoutMessage, setHideTimeoutMessage
 *  1.0.5.1- 2009-10-26- javadoc warnings fixed; remove unused import EOFException
 *  1.2.0 - 2017-06-03 - {@link #setData(String)} now takes a String, not Object.
 *</PRE>
 *
 * @author Jeremy D Monin <jeremy@nand.net>
 * @version 1.2.0
 */
public interface StringConnection
{

    /**
     * @return Hostname of the remote end of the connection
     */
    public abstract String host();

    /**
     * Send data over the connection.
     *<P>
     * <B>Threads:</B> Each implementation must be safe to call from any thread,
     * and synchronize itself on an appropriate object or field.
     *
     * @param str Data to send
     *
     * @throws IllegalStateException if not yet accepted by server
     */
    public abstract void put(String str)
        throws IllegalStateException;

    /** For server-side thread which reads and treats incoming messages */
    public abstract void run();

    /** Are we currently connected and active? */
    public abstract boolean isConnected();

    /** Start ability to read from the net; called only by the server.
     * (In a network-based subclass, another thread may be started by this method.)
     * 
     * @return true if able to connect, false if an error occurred.
     */    
    public abstract boolean connect(); 

    /** Close the socket, set EOF; called after conn is removed from server structures */
    public abstract void disconnect();

    /**
     * Accept no further input, allow output to drain, don't immediately close the socket. 
     * Once called, {@link #isConnected()} will return false, even if output is still being
     * sent to the other side. 
     */
    public abstract void disconnectSoft();

    /**
     * The optional name key used to name this connection.
     *<P>
     * Before v1.2.0, this returned an {@link Object}; getData is a {@link String} in v1.2.0 and up.
     *
     * @return The name key for this connection, or null.
     * @see #getAppData()
     */
    public abstract String getData();

    /**
     * The optional app-specific changeable data for this connection.
     * Not used anywhere in the generic server, only in your app.
     *
     * @return The app-specific data for this connection.
     * @see #getData()
     */
    public abstract Object getAppData();

    /**
     * Set the optional name key for this connection.
     *
     * The StringConnection system uses this data to name the connection,
     * so it should not change once set.
     *<P>
     * If you call setData after {@link Server#newConnection1(StringConnection)},
     * please call {@link Server#nameConnection(StringConnection, boolean)} afterwards
     * to ensure the name is tracked properly at the server.
     *<P>
     * For anything else your application wants to associate with the connection,
     * see {@link #setAppData(Object)}.
     * @param data The new name key, or null
     */
    public abstract void setData(String data);

    /**
     * Set the app-specific non-key data for this connection.
     *
     * This is anything your application wants to associate with the connection.
     * The StringConnection system itself does not reference or use this data.
     * You can change it as often as you'd like, or not use it.
     *
     * @param data The new data, or null
     * @see #setData(String)
     */
    public abstract void setAppData(Object data);

    /**
     * @return Any error encountered, or null
     */
    public abstract Exception getError();

    /**
     * @return Time of connection to server, or of object creation if that time's not available
     *
     * @see #connect()
     */
    public abstract Date getConnectTime();

    /**
     * Give the version number (if known) of the remote end of this connection.
     * The meaning of this number is application-defined.
     * @return Version number, or 0 if unknown.
     */
    public abstract int getVersion();

    /**
     * Set the version number of the remote end of this connection.
     * The meaning of this number is application-defined.
     *<P>
     * <b>Locking:</b> If we're on server side, and {@link #setVersionTracking(boolean)} is true,
     *  caller should synchronize on {@link Server#unnamedConns}.
     *
     * @param version Version number, or 0 if unknown.
     *                If version is greater than 0, future calls to {@link #isVersionKnown()}
     *                should return true.
     */
    public abstract void setVersion(int version);

    /**
     * Set the version number of the remote end of this connection.
     * The meaning of this number is application-defined.
     *<P>
     * <b>Locking:</b> If we're on server side, and {@link #setVersionTracking(boolean)} is true,
     *  caller should synchronize on {@link Server#unnamedConns}.
     *
     * @param version Version number, or 0 if unknown.
     * @param isKnown Should this version be considered confirmed/known by {@link #isVersionKnown()}?
     * @since 1.0.5
     */
    public abstract void setVersion(int version, boolean isKnown);

    /**
     * Is the version known of the remote end of this connection?
     * We may have just assumed it, or taken a default.
     * To confirm, call {@link #setVersion(int, boolean)}.
     * @return True if we've confirmed the version, false if it's assumed or default.
     * @since 1.0.5
     */
    public abstract boolean isVersionKnown();

    /**
     * For server-side use, should we notify the server when our version
     * is changed by setVersion calls?
     * @param doTracking true if we should notify server, false otherwise.
     *        If true, please call both setVersion and
     *        {@link Server#clientVersionAdd(int)} before calling setVersionTracking.
     *        If false, please call {@link Server#clientVersionRem(int)} before
     *        calling setVersionTracking.
     * @since 1.0.5
     */
    public abstract void setVersionTracking(boolean doTracking);

    /**
     * Is input available now, without blocking?
     * Same idea as {@link java.io.DataInputStream#available()}.
     * @since 1.0.5
     */
    public abstract boolean isInputAvailable();

    /**
     * If client connection times out at server, should the server not print a message to console?
     * This would be desired, for instance, in automated clients, which would reconnect
     * if they become disconnected.
     * @see #setHideTimeoutMessage(boolean)
     * @since 1.0.5
     */
    public abstract boolean wantsHideTimeoutMessage();

    /**
     * If client connection times out at server, should the server not print a message to console?
     * This would be desired, for instance, in automated clients, which would reconnect
     * if they become disconnected.
     * @param wantsHide true to hide, false to print, the log message on idle-disconnect
     * @see #wantsHideTimeoutMessage()
     * @since 1.0.5
     */
    public abstract void setHideTimeoutMessage(boolean wantsHide);

}
